#include "sdlab.h"

void uimat_add(uint16_t **dst, uint16_t **src0, uint16_t **src1, int m, int n);
void uimat_expand(uint16_t **to, uint16_t **from, int m, int n);
void uimat_shift(int bits, uint16_t **table, int m, int n);
void uimat_conjunct(uint16_t **dst, uint16_t **src01, uint16_t **src11,
                   uint16_t **src00, uint16_t **src10, int m, int n);
uint16_t** uimat_alloc(int n, int m);
void uimat_free(uint16_t** p, int n, int m);
void uimat_copy(uint16_t *to[], uint16_t *from[], int m, int n);
void uimat_left2right(uint16_t **table, int m, int n);
void uimat_up2down(uint16_t **table, int m, int n);

void print_bits(uint16_t bits, int size);
void print_qam_label(uint16_t **table, int size);



int alloc_count = 0;

void uimat_add(uint16_t **dst, uint16_t **src0, uint16_t **src1, int m, int n)
{
  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      dst[i][j] = src0[i][j] + src1[i][j];
    }
  }
}

void uimat_expand(uint16_t **to, uint16_t **from, int m, int n)
{
  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      to[i][j] = from[i / 2][j / 2];
    }
  }
}

void uimat_shift(int bits, uint16_t **table, int m, int n)
{
  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      table[i][j] <<= bits;
    }
  }
}


void uimat_conjunct(uint16_t **dst, uint16_t **src01, uint16_t **src11,
                         uint16_t **src00, uint16_t **src10, int m, int n)
{
  int pre_m = m / 2;
  int pre_n = n / 2;

  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      if(i < pre_m && j < pre_n){
        dst[i][j] = src01[i][j];
      }else if(i < pre_m && j >= pre_n){
        dst[i][j] = src11[i][j - pre_n];
      }else if(i >= pre_m && j < pre_n){
        dst[i][j] = src00[i - pre_m][j];
      }else if(i >= pre_m && j >= pre_n){
        dst[i][j] = src10[i - pre_m][j - pre_n];
      }
    }
  }
}


uint16_t** uimat_alloc(int n, int m)
{
  uint16_t **ret = (uint16_t**)malloc(n * sizeof(uint16_t*));
  alloc_count++;
  for(int i = 0; i < n; i++){
    ret[i] = (uint16_t*)malloc(m * sizeof(uint16_t));
    alloc_count++;
  }

  return ret;
}

void uimat_free(uint16_t** p, int n, int m)
{
  for(int i = 0; i < n; i++){
    free(p[i]);
    alloc_count--;
  }

  free(p);
  alloc_count--;
}

void uimat_copy(uint16_t *to[], uint16_t *from[], int m, int n)
{
  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      to[i][j] = from[i][j];
    }
  }
}

void uimat_left2right(uint16_t **table, int m, int n)
{
  uint16_t **tmp = uimat_alloc(m, n);
  uimat_copy(tmp, table, m, n);

  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      table[i][j] = tmp[i][n - j - 1];
    }
  }

  uimat_free(tmp, m, n);
}

void uimat_up2down(uint16_t **table, int m, int n)
{
  uint16_t **tmp = uimat_alloc(m, n);
  uimat_copy(tmp, table, m, n);

  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      table[i][j] = tmp[m - i - 1][j];
    }
  }
  uimat_free(tmp, m, n);
}

void print_bits(uint16_t bits, int size)
{
  for(int i = 0; i < size; i++){
    printf("%d", (bits >> (size - i - 1)) & 0x0001);
  }
}

void print_qam_label(uint16_t **table, int size)
{
  int num_col = (int)round(sqrt(size));
  int num_row = num_col;
  int num_bits = num_col;

  for(int i = 0; i < num_col; i++){
    for(int j = 0; j < num_row; j++){
      print_bits(table[i][j], num_col);
      printf(" ");
    }
    printf("\n");
  }
}


void make_qam(int size, uint16_t **qam)
{
  if(size == 4){
    qam[0][0] = 1;
    qam[0][1] = 0;
    qam[1][0] = 3;
    qam[1][1] = 2;
    return;
  }

  if(size == 0){
    printf("illigal input\n");
    exit(-1);
  }

  int n = (int) round(sqrt(size));
  int m = n;
  int pre_n = n / 2;
  int pre_m = m / 2;

  uint16_t **table00 = uimat_alloc(pre_n, pre_m);
  uint16_t **table01 = uimat_alloc(pre_n, pre_m);
  uint16_t **table10 = uimat_alloc(pre_n, pre_m);
  uint16_t **table11 = uimat_alloc(pre_n, pre_m);

  make_qam(pre_n * pre_m, table11);

  uimat_copy(table01, table11, pre_n, pre_m);
  uimat_left2right(table01, pre_n, pre_m);
  uimat_copy(table00, table01, pre_n, pre_m);
  uimat_up2down(table00, pre_n, pre_m);
  uimat_copy(table10, table00, pre_n, pre_m);
  uimat_left2right(table10, pre_n, pre_m);

  uint16_t **lower =uimat_alloc(n, m);

  uimat_conjunct(lower, table01, table11, table00, table10, n, m);

  uint16_t **upper =uimat_alloc(n, m);
  uimat_expand(upper, table11, n, m);

  uimat_shift(pre_n, upper, n, m);
  uimat_add(qam, upper, lower, n, m);

  uimat_free(table00, pre_n, pre_m);
  uimat_free(table01, pre_n, pre_m);
  uimat_free(table10, pre_n, pre_m);
  uimat_free(table11, pre_n, pre_m);
  uimat_free(upper, n, m);
  uimat_free(lower, n, m);
}

int main(int argc, char** argv)
{
  if(argc != 2){
    printf("usage: %s [4096|1024|256|64|16|4]\n", argv[0]);
    exit(-1);
  }

  uint16_t size = atoi(argv[1]);
  uint16_t n = (int)round(sqrt(size));

  uint16_t **qam = uimat_alloc(n, n);
  make_qam(size, qam);
  printf("%d qam\n", size);
  print_qam_label(qam, size);

  uimat_free(qam, n, n);
}
